#!/bin/sh
# \
exec wish -f "$0" ${1+"$@"}

# Client application for Amateur Radio Field Day logging program
# Uses sockets to communicate with remote/local database server

#=======================================================================
# support code for tcl/tk socket services
#=======================================================================

namespace eval ServerSocket {
  namespace export ServerSocket SSclose SSactivate SSwrite SSbroadcast SSdrop

  variable _socket
  variable _clients
  variable _serverFunction ""
}

proc ServerSocket::ServerSocket {{port_ 20002}} {
  variable _socket
  # Open the physical socket.
  if {[catch {socket -server ServerSocket::accept $port_} _socket]} {
    return -code error
  }
  return -code ok
}

proc ServerSocket::SSdrop {sd} {
variable _clients
  unset _clients($sd)
  catch {puts $sd "close"}
  catch {close $sd}
}

proc ServerSocket::SSclose {} {
  variable _socket
  variable _clients
  variable _serverFunction

  catch {close $_socket}
  foreach sock [array names _clients] {
    ServerSocket::SSdrop $sock
  }
  set _serverFunction ""
}

proc ServerSocket::SSactivate {command_} {
variable _serverFunction
  set _serverFunction $command_
}

proc ServerSocket::SSwrite {sd_ data_} {
  if {[catch {puts $sd_ $data_} err]} {
    error $err
  }
}

proc ServerSocket::SSbroadcast {data_} {
  variable _clients
  foreach sock [array names _clients] {
    if {[catch {puts $sock $data_} err]} {
      error $err
    }
  }
}

proc ServerSocket::accept {sd_ address_ port_} {
variable _clients
  set _clients($sd_) [list $address_ $port_]
  fconfigure $sd_ -buffering line
  fileevent $sd_ readable [list ServerSocket::handleClientInput $sd_]
}

proc ServerSocket::handleClientInput {sd_} {
  variable _serverFunction
  variable _clients

  catch {gets $sd_ data} rc

  if {$rc < 0} {
    ServerSocket::SSdrop $sd_
    return
  }
  if {[string compare $data "quit"] == 0} {
    ServerSocket::SSclose
  } else {
    if {[llength $_serverFunction]} {
        eval {$_serverFunction $sd_ $data}
    } else {
      ServerSocket::SSbroadcast $data
    }
  }
}

namespace eval ClientSocket {
  namespace export ClientSocket CSwrite CSregister CSclose
  variable _socket
  variable _clientFunction ""
}

proc ClientSocket::CSclose {} {
  variable _socket
  close $_socket
}

proc ClientSocket::CSregister {command_} {
  variable _clientFunction
  set _clientFunction $command_
}

proc ClientSocket::ClientSocket {{host_ localhost} {port_ 20002}} {
  variable _socket
  if {[catch {socket $host_ $port_} _socket]} {
    return -code error "Server DOWN"
  }
  fconfigure $_socket -buffering line
  fileevent $_socket readable ClientSocket::handleServerInput
  return $_socket
}

proc ClientSocket::CSwrite {data_} {
  variable _socket
  puts $_socket $data_
}

proc ClientSocket::handleServerInput {} {
  variable _socket
  variable _clientFunction
  if {[gets $_socket data] < 0} {
    return
  } else {
    if {[string compare $data "close"] == 0} {
      ClientSocket::CSclose
      $_clientFunction "CLOSE"
      return
    }
    if {[llength $_clientFunction]} {
      $_clientFunction $data
    } else {
      puts $data
    }
  }
}
#=======================================================================

namespace import ClientSocket::*

#=======================================================================
# if client-config.fd exists these values are read from the file during init
# for example
#
# address:localhost
# port:20001
#
#=======================================================================

variable staCallsign STACALL
variable staMult 5
variable staClass 1A
variable socket_addr "localhost" #"127.0.0.1"
variable socket_port 20002
variable datafile "newlog.fd"
variable autosave 60

variable opUser ""
variable opMode ""
variable opBand ""
variable selUser ""
variable selMode ""
variable selBand ""

global contacts
global staCallsign
global staClass
global staMult
global cumScore
global entCall
global entSection
global entClass

variable logged_on 0
variable bands {160 80 40 20 17 15 12 10 6 2 220 440}
variable modes {CW DIG PHONE}

proc Dialog_Create {top title args} {
  global dialog
  if [winfo exists $top] {
    switch -- [wm state $top] {
      normal {
        # Raise a buried window
        raise $top
      }
      withdrawn -
      iconic {
        # Open and restore geometry
        wm deiconify $top
        catch {wm geometry $top $dialog(geo,$top)}
      }
    }
    return 0
  } else {
    eval {toplevel $top} $args
    wm title $top $title
    return 1
  }
}

proc Dialog_Wait {top varName {focus {}}} {
  upvar $varName var

  # Poke the variable if the user nukes the window
  bind $top <Destroy> [list set $varName cancel]

  # Grab focus for the dialog
  if {[string length $focus] == 0} {
    set focus $top
  }

  set old [focus -displayof $top]
  focus $focus
  catch {tkwait visibility $top}
  catch {grab $top}

  # Wait for the dialog to complete
  tkwait variable $varName
  catch {grab release $top}
  focus $old
}

proc Dialog_Dismiss {top} {
  global dialog
  # Save current size and position
  catch {
    # window may have been deleted
    set dialog(geo,$top) [wm geometry $top]
    wm withdraw $top
  }
}

proc FD_Client {line} {
  set command [lindex $line 0]
#  puts stdout $command
  switch $command {
    LOGON_DENIED {FD_logon_denied}
    LOGON_OK {FD_logon_ok [lrange $line 1 end]}
    LOGOFF_OK {}
    LOGOFF_NOP {}
    DUP {FD_dup}
    NODUP {}
    ACCEPT {FD_add_accepted}
    REJECT {FD_add_rejected}
    SCORE {FD_disp_score [lindex $line 1]}
    WORKED {FD_score_update [lrange $line 1 end]}
    CLOSE {FD_client_init}
    QUIT {exit}
    default {}
  }
}

proc FD_score_update {data} {
  global contacts
  foreach bm $data {
    set contacts([lindex $bm 0],[lindex $bm 1]) [lindex $bm 2]
  }
}

proc FD_disp_score {data} {
  global cumScore
  set cumScore $data
}

proc FD_Logon_Prompt {} {
  global selBand selMode selUser modes bands
  global loginYES
  set selBand 80
  set selMode CW
  set f .prompt
  if [Dialog_Create $f "Field Day Log On" -borderwidth 5] {
    set bframe [frame $f.bands -relief groove]
    foreach band $bands {
      radiobutton $bframe.$band -variable selBand \
        -text $band -value $band
      pack $bframe.$band -side left
    }
    pack $bframe -side top
    set mframe [frame $f.modes -relief groove]
    foreach mode $modes {
      radiobutton $mframe.a$mode -variable selMode \
        -text $mode -value $mode
      pack $mframe.a$mode -side left
    }
    pack $mframe -side top
    set eframe [frame $f.e]
      label $eframe.lbl -text "Operator Callsign: " -anchor e
      entry $eframe.ent -textvariable selUser
      pack $eframe.lbl $eframe.ent -side left -fill x
    pack $eframe -side top
    frame $f.btn -relief groove
      button $f.btn.ok -text OK -command {set loginYES 1}
      button $f.btn.cancel -text Cancel -command {set loginYES 0}
      pack $f.btn.ok -side left
      pack $f.btn.cancel -side right
    pack $f.btn -side top
  }
  set loginYES 0
  Dialog_Wait $f loginYES $f.e.ent
  Dialog_Dismiss $f
  return $loginYES
}

proc FD_logon {} {
  global opUser opMode opBand logged_on loginYES
  global selBand selMode selUser
  global socket_addr
  global socket_port
  if {$logged_on == 1} return
  FD_Logon_Prompt
  if {$loginYES == 1} {
    set selUser [string toupper $selUser]
    set selUser [string trim $selUser]
    if {[string length $selUser] == 0} return
    if {[catch {ClientSocket $socket_addr $socket_port } result]} {
      tk_messageBox -type ok -message "Cannot connect to Field Day Server\n$result" -icon error
      exit
    }
    CSregister FD_Client
    CSwrite "LOGON $selUser $selBand,$selMode"
  }
}

proc FD_logon_ok {data} {
  global opUser opMode opBand
  global selUser selMode selBand
  global staCallsign staClass staMult
  global wdgLogOnOff logged_on

  set opMode $selMode
  set opBand $selBand
  set opUser $selUser
  set staCallsign [lindex $data 0]
  set staClass [lindex $data 1]
  set staMult [lindex $data 2]
  set logged_on 1
  $wdgLogOnOff config -text "Log Off"
  $wdgLogOnOff config -command FD_logoff
}

proc FD_logon_denied {} {
  global selBand selMode
  tk_messageBox -type ok -message "$selBand $selMode already logged on" \
     -icon error
#  puts stdout "Logon denied"
}

proc FD_logoff {} {
  global wdgLogOnOff logged_on
  CSwrite "LOGOFF"
  $wdgLogOnOff config -text "Log On"
  $wdgLogOnOff config -command FD_logon
#  puts stdout "FD logoff"
}

proc FD_dupcheck {} {
  global entCall
  set entCall [string toupper $entCall]
  CSwrite "DUPCHECK $entCall"
}

proc FD_dup {} {
  global wdgCall entCall
  $wdgCall config -bg red
  $wdgCall config -fg white
  after 1000 {
    $wdgCall config -bg white
    $wdgCall config -fg black
  }
  set entCall ""
  focus $wdgCall
}

proc FD_add {} {
  global entCall entSection entClass
  if {[string length $entCall] == 0 ||
      [string length $entSection] == 0 ||
      [string length $entClass] == 0 } {
    tk_messageBox -type ok -message "Missing or bad entry!" \
      -icon error
    return
  }
  set entCall [string toupper $entCall]
  set entSection [string toupper $entSection]
  set entClass [string toupper $entClass]
  CSwrite "ADD $entCall $entSection $entClass"
}

proc FD_add_rejected {} {
  global entSection entCall entClass
  global wdgCall wdgSection wdgClass
  set entSection ""
  set entCall ""
  set entClass ""
  $wdgCall config -bg red
  $wdgSection config -bg red
  $wdgClass config -bg red
  after 1000 {
    $wdgCall config -bg white
    $wdgSection config -bg white
    $wdgClass config -bg white
  }
  focus $wdgCall
}

proc FD_add_accepted {} {
  global entSection entCall entClass
  global wdgCall wdgSection wdgClass
  set entSection ""
  set entCall ""
  set entClass ""
  $wdgCall config -bg green
  $wdgSection config -bg green
  $wdgClass config -bg green
  after 1000 {
    $wdgCall config -bg white
    $wdgSection config -bg white
    $wdgClass config -bg white
  }
  focus $wdgCall
}

proc FD_read_config {} {
  global socket_addr socket_port

  if [catch {open "client-config.fd" r} fid] {
#    logentry "Cannot open config.fd\n"
  } else {
    set tmp [read $fid]
    close $fid
    set config [split $tmp "\n"]
    foreach rec $config {
      set lrec [split $rec ":"]
      set name "[lindex $lrec 0]"
      switch $name {
        address    { set socket_addr "[lindex $lrec 1]" }
        port       { set socket_port [lindex $lrec 1] }
        default    {}
      }
    }
  }
}

proc FD_write_config {} {
  global socket_addr socket_port
  if [catch {open "client-config.fd" w} fid] {
    FD_display "Cannot write to config.fd"
  } else {
	puts $fid "address:$socket_addr"
	puts $fid "port:$socket_port"
    close $fid
  }
}

proc FD_client_init {} {
  global contacts bands modes opUser opMode opBand logged_on
  global selUser selMode selBand
  global staCallsign staClass staMult cumScore
  set opUser ""
  set opMode "LOGGED OFF"
  set opBand ""
  set selUser ""
  set selMode ""
  set selBand ""
  set staCallsign ""
  set staClass ""
  set staMult ""
  set cumScore ""
  set logged_on 0
  foreach band $bands {
    foreach mode $modes {
      set contacts($band,$mode) 0
    }
  }
  FD_read_config
}

proc FD_client_exit {} {
#  puts stdout "FD client exit"
  global logged_on
  if {$logged_on == 1} {
    FD_logoff
  }
  exit
}

# Procedures to help build dialogs.
#

proc Dialog_Create {top title args} {
	global dialog
	if [winfo exists $top] {
		switch -- [wm state $top] {
			normal {
				# Raise a buried window
				raise $top
			}
			withdrawn -
			iconic {
				# Open and restore geometry
				wm deiconify $top
				catch {wm geometry $top $dialog(geo,$top)}
			}
		}
		return 0
	} else {
		eval {toplevel $top} $args
		wm title $top $title
		return 1
	}
}

proc Dialog_Wait {top varName {focus {}}} {
	upvar $varName var

	# Poke the variable if the user nukes the window
	bind $top <Destroy> [list set $varName cancel]

	# Grab focus for the dialog
	if {[string length $focus] == 0} {
		set focus $top
	}
	set old [focus -displayof $top]
	focus $focus
	catch {tkwait visibility $top}
	catch {grab $top}

	# Wait for the dialog to complete
	tkwait variable $varName
	catch {grab release $top}
	focus $old
}

proc Dialog_Dismiss {top} {
	global dialog
	# Save current size and position
	catch {
		# window may have been deleted
		set dialog(geo,$top) [wm geometry $top]
		wm withdraw $top
	}
}

proc FCClientConfigDialog {} {
	global prompt socket_addr socket_port
	set f .dialog
	set prompt(Address) $socket_addr
	set prompt(Port) $socket_port
	if [Dialog_Create $f "TCP/IP config" -borderwidth 10] {
		frame $f.entries
		foreach field {Address Port} {
			label $f.entries.l$field -text $field -anchor w
			entry $f.entries.e$field -textvariable prompt($field) \
			  -relief sunken -bg white
			grid $f.entries.l$field $f.entries.e$field -sticky news
		}
		message $f.msg -text "---------------------------------" -aspect 1000
		set b [frame $f.buttons]
		pack $f.entries $f.msg $f.buttons -side top -fill x
		button $b.ok -text OK -command {set prompt(ok) 1}
		button $b.cancel -text Cancel -command {set prompt(ok) 0}
		pack $b.ok -side left
		pack $b.cancel -side right
		bind $f.entries.eAddress <Return> {set prompt(ok) 1 ; break}
		bind $f.entries.eAddress <Control-c> {set prompt(ok) 0 ; break;}
		bind $f.entries.ePort <Return> {set prompt(ok) 1 ; break}
		bind $f.entries.ePort <Control-c> {set prompt(ok) 0 ; break;}
	}
	set prompt(ok) 0
	Dialog_Wait $f prompt(ok)
	Dialog_Dismiss $f
	if {$prompt(ok)} {
		set socket_addr $prompt(Address)
		set socket_port $prompt(Port)
		FD_client_init
		FD_write_config
	}
}

proc FD_About {} {
  tk_messageBox -type ok \
     -message "Created by W1HKJ" \
     -icon info
}

proc FDClientMainWindow {} {
global contacts modes bands
global opMode opBand opUser
global staCallsign staClass staMult
global entCall entSection entClass
global wdgCall wdgSection wdgClass
global cumScore
global wdgLogOnOff

  wm withdraw .
  wm title . "Field Day Client"
  wm protocol . WM_DELETE_WINDOW FD_client_exit

  menu .menubar
  . config -menu .menubar
  foreach m {File Config View Help} {
    set $m [menu .menubar.m$m -tearoff 0]
    .menubar add cascade -label $m -menu .menubar.m$m
  }
  $File add command -label Exit -command FD_client_exit
  $Config add command -label Config -command FCClientConfigDialog
  $Help add command -label About -command FD_About

  frame .f -borderwidth 4
    label .f.l00 -text "Sta Call:" -anchor e
    label .f.l01 -textvariable staCallsign
    label .f.l10 -text "    Class:" -anchor e
    label .f.l11 -textvariable staClass
    label .f.l20 -text "    Mult:" -anchor e
    label .f.l21 -textvariable staMult
    pack .f.l00 -side left -fill x -anchor e -expand true
    pack .f.l01 -side left -anchor w
    pack .f.l10 -side left -anchor e
    pack .f.l11 -side left -anchor w
    pack .f.l20 -side left -anchor e
    pack .f.l21 -side left -anchor w -fill x -expand true
  pack .f -side top

  frame .f5
    label .f5.l1 -text "Cumulative Score:"
    label .f5.l2 -textvariable cumScore
    pack .f5.l1 .f5.l2 -side left -fill x -expand true
  pack .f5 -side top


  frame .f0 -borderwidth 4
    set r 0
    foreach mode $modes {
      incr r
      label .f0.m$mode -text $mode -anchor e -relief flat
      grid .f0.m$mode -sticky news -row $r -column 0
    }
    set c 0
    foreach band $bands {
      incr c
      label .f0.b$band -text $band -relief flat
      grid .f0.b$band -sticky news -row 0 -column $c
    }
    set r 0
    foreach mode $modes {
      incr r
      set c 0
      foreach band $bands {
        incr c
        label .f0.$band$mode \
              -textvariable contacts($band,$mode) \
              -width 5 -height 1 -bg white \
              -relief sunken \
              -anchor e
        grid .f0.$band$mode -sticky news -row $r -column $c
      }
    }
  pack .f0 -side top

  frame .f1
    label .f1.band \
      -textvariable opBand \
      -width 15 -relief flat
    label .f1.mode \
      -textvariable opMode \
      -width 15 -relief flat
    label .f1.opUser \
      -textvariable opUser \
      -width 15 -relief flat
    pack .f1.band .f1.mode .f1.opUser -side left -fill x -expand true
  pack .f1 -side top -fill x -expand true

  frame .f2 -borderwidth 4
    label .f2.lb1 -text "CALLSIGN"
    label .f2.lb2 -text "CLASS"
    label .f2.lb3 -text "SECTION"
    set wdgCall [entry .f2.call \
      -textvariable entCall \
      -width 8 \
      -bg white ]
    set wdgSection [entry .f2.class \
      -textvariable entClass \
      -width 5 \
      -bg white ]
    set wdgClass [entry .f2.section \
      -textvariable entSection \
      -width 8 \
      -bg white ]
    label .f2.lb4 -text "    "
    label .f2.lb5 -text "    "
    grid .f2.lb1 -sticky news -row 0 -column 0
    grid .f2.lb2 -sticky news -row 0 -column 2
    grid .f2.lb3 -sticky news -row 0 -column 4
    grid .f2.call -sticky news -row 1 -column 0
    grid .f2.lb4 -row 1 -column 1
    grid .f2.class -sticky news -row 1 -column 2
    grid .f2.lb5 -row 1 -column 3
    grid .f2.section -sticky news -row 1 -column 4

    label .f2.l5 -text "    "
    label .f2.l10 -text "          "
    label .f2.l20 -text "                          "
    button .f2.b0 -text "Dup Check" -command FD_dupcheck
    button .f2.b1 -text "Enter" -command FD_add
    set wdgLogOnOff [button .f2.b2 -text "LogOn" -command FD_logon]

    grid .f2.l5 -sticky news -row 1 -column 5
    grid .f2.b0 -sticky news -row 1 -column 6
    grid .f2.l10 -sticky news -row 1 -column 7
    grid .f2.b1 -sticky news -row 1 -column 8
    grid .f2.l20 -sticky news -row 1 -column 9
    grid .f2.b2 -sticky news -row 1 -column 10
    
  pack .f2 -fill x -expand true

}

#================================================================

proc main {} {

  FDClientMainWindow

  wm deiconify .

  FD_client_init
}

wm iconify .

main
